package higa.jhcl.control;

import java.awt.Insets;

import java.text.Format;

import higa.jhcl.model.ItemPainter;
import higa.jhcl.model.ItemEditor;
import higa.jhcl.model.SingleMatrixSelection;
import higa.jhcl.model.BasicMatrixSelection;
import higa.jhcl.model.BasicMatrixContainer;
import higa.jhcl.model.VectorModel;
import higa.jhcl.model.VectorModelEvent;
import higa.jhcl.model.VectorModelListener;
import higa.jhcl.model.MatrixModel;
import higa.jhcl.model.MatrixModelEvent;
import higa.jhcl.model.MatrixModelListener;
import higa.jhcl.model.MatrixLocation;

import higa.jhcl.view.ButtonItemPainter;
import higa.jhcl.view.FocusableItemPainter;
import higa.jhcl.view.SelectableItemPainter;
import higa.jhcl.view.TextItemPainter;
import higa.jhcl.view.TextItemEditor;
import higa.jhcl.view.ChoiceItemPainterEditor;
import higa.jhcl.view.GridView;
import higa.jhcl.view.ColumnViewManager;
import higa.jhcl.view.ColumnView;

import higa.sql.dataset.IDataSet;
import higa.sql.dataset.DataSetAware;
import higa.sql.dataset.Column;
import higa.sql.dataset.NavigationEvent;
import higa.sql.dataset.NavigationListener;
import higa.sql.dataset.DataChangeEvent;
import higa.sql.dataset.DataChangeListener;
import higa.sql.dataset.MatrixDataSetManager;

import higa.util.EventMulticaster;
import higa.util.Alignment;
import higa.util.JDBCUtil;
import higa.util.Formatter;

public class HGrid extends GridView
  implements DataSetAware,
  NavigationListener, DataChangeListener {

  private IDataSet dataSet;
  private boolean multiSelect = false;

  private transient boolean addNotifyCalled = false;

  public HGrid() {
    resetSelection();
    buildStringGrid(null);

    ItemPainter painter = new FocusableItemPainter(
      new SelectableItemPainter(new TextItemPainter()));

    setPainter(painter);

    ItemEditor editor = new TextItemEditor();
    setEditor(editor);
  }

  public void setDataSet(IDataSet ds) {
    if (dataSet != null) {
      dataSet.removeDataChangeListener(this);
      dataSet.removeNavigationListener(this);
    }

    dataSet = ds;
    if (dataSet != null && addNotifyCalled) {
      openDataSet();
    } else {
      buildStringGrid(null);
      int count = getColumnCount();
      for (int i = 0; i < count; i++) {
        ColumnView cv = getDefaultColumnView(i);
        setColumnView(i, cv);
      }
    }

    if (dataSet != null) {
      dataSet.addDataChangeListener(this);
      dataSet.addNavigationListener(this);
    }
  }

  public IDataSet getDataSet() {
    return dataSet;
  }

  public synchronized void setItems(String[][] newItems) {
    if (dataSet != null) {
      return;
    }
    
    buildStringGrid(newItems);
  }

  public String[][] getItems() {
    MatrixModel model = getModel();
    if (model == null ||
      model instanceof MatrixDataSetManager) {

      return new String[0][0];
    }

    int rowCount = model.getRowCount();
    int columnCount = model.getColumnCount();
    String[][] items = new String[rowCount][columnCount];
    Format[] formatters = getColumnFormatters();
    for (int r = 0; r < rowCount; r++) {
      for (int c = 0; c < columnCount; c++) {
        Object data = model.get(r, c);
        items[r][c] = Formatter.format(
          formatters[c], data);
      }
    }
    
    return items;
  }

  public void addNotify() {
    super.addNotify();
    if (!addNotifyCalled) {
      addNotifyCalled = true;
      if (dataSet != null) {
        openDataSet();
      }
    }
  }

  public void setColumnCaptions(String[] captions) {
    if (captions == null) {
      return;
    }

    int count = captions.length;
    int columnCount = getColumnCount();
    ColumnViewManager cvm = getColumnViewManager();
    cvm.setCount(count);
    for (int i = 0; i < count; i++) {
      ColumnView cv = getColumnView(i);
      cv.setCaption(captions[i]);
      setColumnView(i, cv);
    }
  }

  public String[] getColumnCaptions() {
    int count = getColumnCount();
    String[] captions = new String[count];
    for (int i = 0; i < count; i++) {
      ColumnView cv = getColumnView(i);
      captions[i] = cv.getCaption();
    }

    return captions;
  }

  public void setColumnFormatPatterns(String[] patterns) {
    if (patterns == null) {
      return;
    }

    int count = patterns.length;
    int columnCount = getColumnCount();
    int minCount = Math.min(count, columnCount);
    ColumnViewManager cvm = getColumnViewManager();
    cvm.setCount(columnCount);
    for (int i = 0; i < minCount; i++) {
      ColumnView cv = getColumnView(i);
      cv.setFormatPattern(patterns[i]);
      setColumnView(i, cv);
    }
  }

  public String[] getColumnFormatPatterns() {
    int count = getColumnCount();
    String[] patterns = new String[count];
    for (int i = 0; i < count; i++) {
      ColumnView cv = getColumnView(i);
      patterns[i] = cv.getFormatPattern();
    }

    return patterns;
  }

  public Format[] getColumnFormatters() {
    int count = getColumnCount();
    Format[] formatters = new Format[count];
    for (int i = 0; i < count; i++) {
      ColumnView cv = getColumnView(i);
      formatters[i] = cv.getFormatter();
    }

    return formatters;
  }

  public void dataChanged(DataChangeEvent e) {
    adjustSubfocus();
  }

  public void navigated(NavigationEvent e) {
    adjustSubfocus();
  }

  private void adjustSubfocus() {
    int rowIndex = dataSet.getRowIndex();
    MatrixLocation ml = getSubfocus();
    int mlRowIndex = ml != null ? ml.rowIndex : 0;
    int mlColumnIndex = ml != null ? ml.columnIndex : 0;
    if (mlRowIndex != rowIndex) {
      setSubfocus(rowIndex, mlColumnIndex);
    }
  }

  public void setMultiSelect(boolean select) {
    if (multiSelect != select) {
      multiSelect = select;
      resetSelection();
    }
  }

  public boolean isMultiSelect() {
    return multiSelect;
  }

  private void resetSelection() {
    if (multiSelect) {
      setSelection(new BasicMatrixSelection(
        getSelection().getAll()));
    } else {
      setSelection(new SingleMatrixSelection(getSubfocus()));
    }

    repaint();
  }

  private void buildStringGrid(String[][] newItems) {
    if (newItems == null) {
      newItems = new String[1][1];
    }

    int columnCount = newItems[0].length;
    int rowCount = newItems.length;
    ColumnViewManager cvm = getColumnViewManager();
    cvm.setCount(columnCount);

    BasicMatrixContainer newModel =
      new BasicMatrixContainer(newItems);

    setModel(newModel);
  }

  private void openDataSet() {
    if (dataSet == null || !addNotifyCalled) {
      return;
    }

    if (!dataSet.isOpen()) {
      try {
        dataSet.open();
      } catch (Exception ex) {
        ex.printStackTrace();
        return;
      }
    }

    if (dataSet.isOpen()) {
      bindDataSet();
    }
  }

  private synchronized void bindDataSet() {
    if (dataSet == null) {
      return;
    }

    setBatchMode(true);

    ColumnViewManager cvm = getColumnViewManager();
    int columnCount = dataSet.getColumnCount();
    cvm.setCount(columnCount);

    for (int c = 0; c < columnCount; c++) {
      Column column = dataSet.getColumn(c);
      ColumnView cv = cvm.getDefaultColumnView(c);
      ColumnView oldCv = cvm.getColumnView(c);
      String name = column.getName();
      cv.setName(name);
      String caption = column.getCaption();
      cv.setCaption(caption);
      int type = column.getType();
      cv.setType(type);
      if (JDBCUtil.isNumber(type)) {
        cv.setAlignment(Alignment.RIGHT | Alignment.MIDDLE);
      }

      String pickListTableName = column.getPickListTableName();
      String pickListColumnName = column.getPickListColumnName();
      String pickListDisplayColumnName= column.getPickListDisplayColumnName();
      if (pickListTableName != null &&
        pickListColumnName != null &&
        pickListDisplayColumnName != null) {

        ChoiceItemPainterEditor cipe =
          new ChoiceItemPainterEditor();

        cipe.setPickListTableName(pickListTableName);
        cipe.setPickListColumnName(pickListColumnName);
        cipe.setPickListDisplayColumnName(pickListDisplayColumnName);
        cipe.setColumnName(name);
        cipe.setDataSet(dataSet);
        cipe.setupItems();
        cv.setPainter(cipe);
        cv.setEditor(cipe);
      }

      cv.merge(oldCv);
      cvm.setColumnView(c, cv);
    }

    MatrixDataSetManager model =
      new MatrixDataSetManager(dataSet);

    setModel(model);

    addSubfocusListener(model);
    setBatchMode(false);
  }
}
