package higa.jhcl.view;

import java.io.Serializable;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.IOException;

import java.awt.Insets;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Color;
import java.awt.SystemColor;
import java.awt.Rectangle;

import higa.jhcl.model.ItemPainter;
import higa.jhcl.model.ItemPaintSite;

public class BorderItemPainter implements ItemPainter,
  Serializable {

  public static final int OUTER_RAISED = 0x0100;
  public static final int OUTER_SUNKEN = 0x0200;
  public static final int INNER_RAISED = 0x0400;
  public static final int INNER_SUNKEN = 0x0800;
  public static final int EDGE_RAISED = OUTER_RAISED | INNER_RAISED;
  public static final int EDGE_SUNKEN = OUTER_SUNKEN | INNER_SUNKEN;
  public static final int EDGE_ETCHED = OUTER_SUNKEN | INNER_RAISED;
  public static final int EDGE_BUMP = OUTER_RAISED | INNER_SUNKEN;
  public static final int EDGE_OUTER = OUTER_RAISED | OUTER_SUNKEN;
  public static final int EDGE_INNER = INNER_RAISED | INNER_SUNKEN;

  private int edge;
  private transient ItemPainter painter;

  public BorderItemPainter() {
    this(0);
  }

  public BorderItemPainter(int edge) {
    this.edge = edge;
  }

  public BorderItemPainter(ItemPainter painter) {
    this();
    this.painter = painter;
  }

  public BorderItemPainter(ItemPainter painter, int edge) {
    this(edge);
    this.painter = painter;
  }

  public void setEdge(int edge) {
    this.edge = edge;
  }

  public int getEdge() {
    return edge;
  }

  public void setPainter(ItemPainter painter) {
    this.painter = painter;
  }

  public ItemPainter getPainter() {
    return painter;
  }

  public Insets getInsets() {
    int thickness = getThickness();

    return new Insets(
      thickness, thickness, thickness, thickness);
  }

  public int getThickness() {
    return ((edge & EDGE_OUTER) != 0 ? 1 : 0) +
      ((edge & EDGE_INNER) != 0 ? 1 : 0);
  }

  public Dimension getPreferredSize(Object data,
    Graphics g, int state, ItemPaintSite site) {

    Insets insets = getInsets();
    int width = insets.left + insets.right;
    int height = insets.top + insets.bottom;

    if (painter != null) {
      Dimension innerSize = painter.getPreferredSize(
        data, g, state, site);

      width += innerSize.width;
      height += innerSize.height;
    }

    return new Dimension(width, height);
  }

  public void paint(Object data, Graphics g,
    Rectangle bounds, int state, ItemPaintSite site) {

    Rectangle clip = g.getClipBounds();
    if (clip == null) {
      return;
    }

    g.clipRect(bounds.x, bounds.y, bounds.width, bounds.height);
    drawEdge(g, bounds, edge);

    if (painter != null) {
      Insets insets = getInsets();
      int thickness = getThickness();
      bounds.x += insets.left;
      bounds.width -= insets.left + insets.right;
      bounds.y += insets.top;
      bounds.height -= insets.top + insets.bottom;
      g.setClip(bounds.x, bounds.y, bounds.width, bounds.height);
      painter.paint(data, g, bounds, state, site);
    }

    g.setClip(clip.x, clip.y, clip.width, clip.height);
  }

  public static void drawEdge(
    Graphics g, Rectangle bounds,int edge) {
    Rectangle b = new Rectangle(bounds);

    if ((edge & EDGE_OUTER) != 0) {
      Color[] tlColors = {
        SystemColor.controlHighlight,
        SystemColor.controlShadow
      };

      Color[] brColors = {
        SystemColor.controlDkShadow,
        SystemColor.controlLtHighlight
      };

      int ci = (edge & OUTER_SUNKEN) != 0 ? 1 : 0;

      paintFrame(g, b, tlColors[ci], brColors[ci]);
      b.grow(-1,-1);
    }

    if ((edge & EDGE_INNER) != 0) {
      Color[] tlColors = {
        SystemColor.controlLtHighlight,
        SystemColor.controlDkShadow
      };

      Color[] brColors = {
        SystemColor.controlShadow,
        SystemColor.controlHighlight
      };

      int ci = (edge & INNER_SUNKEN) != 0 ? 1 : 0;
      paintFrame(g, b, tlColors[ci], brColors[ci]);
      b.grow(-1,-1);
    }
  }

  static void paintFrame(Graphics g, Rectangle b,
    Color tlColor, Color brColor) {

    g.setColor(tlColor);
    g.drawLine(b.x, b.y + 1, b.x, b.y + b.height - 1);
    g.drawLine(b.x, b.y, b.x + b.width - 1, b.y);

    g.setColor(brColor);
    g.drawLine(b.x + b.width - 1, b.y,
      b.x + b.width - 1, b.y + b.height - 1);

    g.drawLine(b.x, b.y + b.height - 1,
      b.x + b.width - 1, b.y + b.height - 1);
  }

  private void writeObject(ObjectOutputStream s)
    throws IOException {

    s.defaultWriteObject();
    s.writeObject(
      painter instanceof Serializable ? painter : null);
  }

  private void readObject(ObjectInputStream s)
    throws IOException, ClassNotFoundException {
    
    s.defaultReadObject();
    Object data = s.readObject();
    if (data instanceof ItemPainter) {
      painter = (ItemPainter)data;
    }
  }
}