package higa.sql.dataset;

import java.io.Serializable;                              

import java.util.Hashtable;
import java.util.Vector;
import java.util.Stack;
import java.util.StringTokenizer;
import java.util.EmptyStackException;

import java.text.Format;

import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;

import higa.util.EventMulticaster;
import higa.util.JDBCUtil;
import higa.util.StringUtil;
import higa.util.Formatter;
import higa.util.VetoException;

public class HDataSet implements IDataSet,
  TransactionListener, MasterDataChangeListener,
  Serializable, Cloneable {

  private HDatabase database;
  private String query;
  private String updateTableName;
  private IDataSet masterDataSet;
  private String[] masterColumnNames = new String[0];
  private String[] detailColumnNames = new String[0];
  private int fetchSize = 20;
  private boolean stringPrepare = true;

  private Vector rows = new Vector();
  private int currentRowIndex = -1;

  private transient LogEntry logEntry;
  private transient Stack log = new Stack();

  private Hashtable parameters = new Hashtable();

  private transient Object[] masterValues = new Object[0];
  private transient int[] masterTypes = new int[0];
  private transient int oldMasterRowid = -1;
  private transient Hashtable cashe = new Hashtable();
  private transient boolean detailCashe = true;

  private ColumnManager columnManager;
  private String catalogName;
  private String schemaName;
  private String tableName;
  private String[] whereColumns = new String[0];

  private EventMulticaster dataChangeCaster =
    new EventMulticaster();

  private EventMulticaster navigationCaster =
    new EventMulticaster();

  private EventMulticaster masterDataChangeCaster =
    new EventMulticaster();

  private EventMulticaster statusCaster =
    new EventMulticaster();

  private transient PreparedStatement
    selectStatement;

  private transient boolean traceSQL = false;
  private transient ResultSet resultSet;
  private transient boolean fetchNeed = true;
  private boolean metaDataNeed = true;

  public HDataSet() {}

  public HDataSet(HDataSet ds) {
    setDataSet(ds);
  }

  public Object clone() {
    return new HDataSet(this);
  }

  public synchronized void setDatabase(
    HDatabase database) throws SQLException {

    if (this.database != null) {
      this.database.removeTransactionListener(this);
    }

    this.database = database;
    if (database != null) {
      database.addTransactionListener(this);
    }

    _close();
  }

  public HDatabase getDatabase() {
    return database;
  }

  public synchronized void setQuery(
    String query) throws SQLException {

    this.query = query;
    _close();
  }

  public String getQuery() {
    return query;
  }

  public void setUpdateTableName(
    String updateTableName) {

    this.updateTableName = updateTableName;
  }

  public String getUpdateTableName() {
    return updateTableName;
  }

  public synchronized void setMasterDataSet(
    IDataSet ds) {

    if (masterDataSet != null) {
      masterDataSet.removeMasterDataChangeListener(this);
    }

    masterDataSet = ds;
    if (masterDataSet != null) {
      masterDataSet.addMasterDataChangeListener(this);
    }
  }

  public IDataSet getMasterDataSet() {
    return masterDataSet;
  }

  public synchronized void setMasterColumnNames(
    String[] columnNames) {

    if (columnNames == null ||
      columnNames.length == 0) {

      masterColumnNames = null;
    } else {
      masterColumnNames =
        StringUtil.toUpperStringArray(columnNames);
    }
  }

  public String[] getMasterColumnNames() {
    return masterColumnNames;
  }

  public synchronized void setDetailColumnNames(
    String[] columnNames) {

    if (columnNames == null ||
      columnNames.length == 0) {

      detailColumnNames = null;
    } else {
      detailColumnNames =
        StringUtil.toUpperStringArray(columnNames);
    }
  }

  public String[] getDetailColumnNames() {
    return detailColumnNames;
  }

  public synchronized void setFetchSize(
    int fetchSize) {

    this.fetchSize = fetchSize;
  }

  public int getFetchSize() {
    return fetchSize;
  }

  public synchronized void setStringPrepare(
    boolean stringPrepare) {

    this.stringPrepare = stringPrepare;
  }

  public boolean isStringPrepare() {
    return stringPrepare;
  }

  public synchronized void setTraceSQL(
    boolean traceSQL) {

    this.traceSQL = traceSQL;
  }

  public boolean isTraceSQL() {
    return traceSQL;
  }

  public synchronized void open()
    throws SQLException {

    if (database == null || query == null) {
      return;
    }

    close();

    if (traceSQL) {
      System.out.println(query);
    }

    selectStatement =
      database.prepareStatement(query);

    executeQuery();
    setupMetaData();
    first();
  }

  public boolean isOpen() {
    return resultSet != null;
  }

  public synchronized void close()
    throws SQLException {

    commit();
    _close();
  }

  protected synchronized void _close()
    throws SQLException {

    log = new Stack();
    logEntry = null;

    __close();
  }

  protected synchronized void __close()
    throws SQLException {

    if (resultSet != null) {
      resultSet.close();
      resultSet = null;
    }

    fetchNeed = false;
    metaDataNeed = true;

    if (selectStatement != null) {
      selectStatement.close();
      selectStatement = null;
    }

    rows.removeAllElements();
    currentRowIndex = -1;
  }

  private synchronized void setupMetaData()
    throws SQLException {

    if (!metaDataNeed) {
      return;
    }

    ResultSetMetaData rsmd =
      resultSet.getMetaData();

    columnManager = new ColumnManager(rsmd);

    catalogName = rsmd.getCatalogName(1);
    schemaName = rsmd.getSchemaName(1);
    tableName = rsmd.getTableName(1);

    if (schemaName == null ||
      schemaName.length() == 0) {

      schemaName = database.getUserName();
    }

    if (tableName == null||
      tableName.length() == 0) {

      String s = " FROM ";
      int startPos = query.toUpperCase().indexOf(s) +
        s.length();

      if (startPos > 0) {
        tableName = query.substring(startPos).trim();
        StringTokenizer st =
          new StringTokenizer(tableName, " \n\t\r,");

        if (st.hasMoreTokens()) {
          tableName = st.nextToken();
        }
      }
    }

    Vector v = new Vector();
    try {
      ResultSet rsPkey =
        database.getMetaData().getPrimaryKeys(
        catalogName, schemaName, tableName);


      while (rsPkey.next()) {
        v.addElement(
          rsPkey.getString(4).toUpperCase());
      }
    } catch (Exception ex) {}

    int count = v.size();
    if (count > 0) {
      whereColumns = new String[count];
      for (int i = 0; i < count; i++) {
        whereColumns[i] =
          (String)v.elementAt(i);
      }
    } else {
      count = getColumnCount();
      for (int i = 0; i < count; i++) {
        Column column = columnManager.getColumn(i);
        if (!column.isNullable()) {
          v.addElement(getColumnName(i));
        }
      }

      count = v.size();
      if (count > 0) {
        whereColumns = new String[count];
        for (int i = 0; i < count; i++) {
          whereColumns[i] =
            (String)v.elementAt(i);
        }
      } else {
        count = getColumnCount();
        whereColumns = new String[count];
        for (int i = 0; i < count; i++) {
          whereColumns[i] = getColumnName(i);
        }
      }
    }

    metaDataNeed = false;
  }
  
  public synchronized void setInParameter(
    int index, Object value) throws Exception {

    parameters.put(new Integer(index), value);
  }

  public synchronized void clearParameters()
    throws SQLException {

    if (selectStatement != null) {
      selectStatement.clearParameters();
    }

    parameters.clear();
  }

  public void insertRow() throws SQLException {
    insertRow(currentRowIndex + 1);
  }

  public void insertRow(int rowIndex) throws SQLException {
    _insertRow(rowIndex);
    fireDataChanged();
    fireMasterDataChanged();
    fireStatusMessage();
  }

  protected synchronized void _insertRow(int rowIndex)
    throws SQLException {
    
    _post();
    logEntry = new LogEntry(
      rowIndex, Status.INSERTED);

    __insertRow(rowIndex);
  }

  protected synchronized void __insertRow(int rowIndex) {
    rows.insertElementAt(new ReadWriteRow(), rowIndex);
    currentRowIndex = rowIndex;
  }

  public void set(String columnName,
    Object value) throws SQLException {

    int columnIndex =
      getColumnIndex(columnName);

    set(columnIndex, value);
  }

  public void set(int columnIndex, Object value)
    throws SQLException {

    set(currentRowIndex, columnIndex, value);
  }

  public void set(int rowIndex, int columnIndex,
    Object value) throws SQLException {

    _set(rowIndex, columnIndex, value);
    fireDataChanged();
    fireStatusMessage();
  }

  protected synchronized void _set(
    int rowIndex, int columnIndex, Object value)
    throws SQLException {

    if (isRowEmpty()) {
      __deleteRow(rowIndex);
      _insertRow(rowIndex);
    } else if (currentRowIndex != rowIndex) {
      _post();
    }

    if (logEntry == null) {
      logEntry = new LogEntry(
        rowIndex, Status.UPDATED,
        new ReadRow((ReadRow)rows.elementAt(rowIndex)));
    }

    logEntry.setColumnValue(columnIndex, value);

    __set(rowIndex, columnIndex, value);
  }

  protected synchronized void __set(
    int rowIndex, int columnIndex, Object value)
    throws SQLException {

    ReadWriteRow rwr =
      (ReadWriteRow)rows.elementAt(rowIndex);

    rwr.set(columnIndex, value);
    currentRowIndex = rowIndex;
  }

  public void deleteRow() throws SQLException {
    deleteRow(currentRowIndex);
  }

  public void deleteRow(int rowIndex)
    throws SQLException {

    _deleteRow(rowIndex);
    fireDataChanged(Status.STRUCTURE_CHANGED);
    fireMasterDataChanged();
    fireStatusMessage();
  }

  protected synchronized void _deleteRow(int rowIndex)
    throws SQLException {

    if (getStatus() == Status.INSERTED) {
      logEntry = null;
    } else {
      ReadRow rr = getRow(rowIndex);
      _post();
      logEntry = new LogEntry(
        rowIndex, Status.DELETED,
        new ReadRow(rr));
    }

    __deleteRow(rowIndex);
  }

  protected synchronized void __deleteRow(int rowIndex) {
    rows.removeElementAt(rowIndex);
    if (rowIndex == getRowCount()) {
      currentRowIndex--;
    }
  }

  protected void setRow(ReadRow rr) {
    setRow(currentRowIndex, rr);
  }

  protected synchronized void setRow(int rowIndex, ReadRow rr) {
    rows.setElementAt(new ReadWriteRow(rr), rowIndex);
    currentRowIndex = rowIndex;
  }

  public void post() throws SQLException {
    if (logEntry == null) {
      return;
    }

    int oldStatus = getStatus();

    _post();

    switch (oldStatus) {
      case Status.INSERTED:
      case Status.DELETED:
        fireMasterDataChanged();
        break;
    }

    //fireDataChanged();
    fireStatusMessage();
  }

  protected synchronized void _post() throws SQLException {
    int status = getStatus();

    switch (status) {
      case Status.INSERTED:
        processInsert();
        break;
      case Status.UPDATED:
        processUpdate();
        break;
      case Status.DELETED:
        processDelete();
        break;
    }

    if (logEntry != null) {
      log.push(logEntry);
      logEntry = null;
    }
  }

  protected synchronized void processInsert()
    throws SQLException {

    ColumnValue[] cvs =
      logEntry.getColumnValues();

    int count = cvs.length;
    if (count == 0) {
      return;
    }

    String tName = updateTableName == null ||
      updateTableName.length() == 0 ?
      tableName : updateTableName;

    StringBuffer sb = new StringBuffer(
      "INSERT INTO " + tName + "(");

    StringBuffer sb2 = new StringBuffer(
      " VALUES(");

    for (int i = 0; i < count; i++) {
      int index = cvs[i].getColumnIndex();
      String columnName =
        getColumnName(index);

      sb.append(columnName + ",");
      Column column = columnManager.getColumn(index);
      if (stringPrepare ||
        !JDBCUtil.isString(column.getType())) {

        sb2.append("?");
      } else {
        sb2.append("'" + get(index) + "'");
      }

      sb2.append(",");
    }

    int dCount = detailColumnNames.length;
    for (int i = 0; i < dCount; i++) {
      if (null == masterValues[i]) {
        continue;
      }

      int index = getColumnIndex(detailColumnNames[i]);
      set(index, masterValues[i]);

      sb.append(detailColumnNames[i] + ",");
      if (stringPrepare ||
        !JDBCUtil.isString(masterTypes[i])) {

        sb2.append("?");
      } else {
        sb2.append("'" + masterValues[i] + "'");
      }

      sb2.append(",");
    }

    sb.setCharAt(sb.length() - 1, ')');
    sb2.setCharAt(sb2.length() - 1, ')');
    sb.append(sb2.toString());

    executeUpdate(sb.toString(), cvs, Status.INSERTED);
  }

  protected synchronized void processUpdate()
    throws SQLException {

    ColumnValue[] cvs =
      logEntry.getColumnValues();

    String tName = updateTableName == null ||
      updateTableName.length() == 0 ?
      tableName : updateTableName;

    StringBuffer sb = new StringBuffer(
      "UPDATE " + tName + " SET ");

    int count = cvs.length;
    for (int i = 0; i < count; i++) {
      int index = cvs[i].getColumnIndex();
      String columnName =
        getColumnName(index);

      Column column = columnManager.getColumn(index);
      if (stringPrepare ||
        !JDBCUtil.isString(column.getType())) {

        sb.append(columnName + " = ?");
      } else {
        sb.append(columnName + " = '" + get(index) + "'");
      }

      sb.append(",");
    }

    sb.setLength(sb.length() - 1);

    executeUpdate(sb.toString(), cvs, Status.UPDATED);
  }

  protected synchronized void processDelete()
    throws SQLException {

    String tName = updateTableName == null ||
      updateTableName.length() == 0?
      tableName : updateTableName;

    executeUpdate(
      "DELETE FROM " + tName,
      new ColumnValue[0], Status.DELETED);
  }

  private void executeUpdate(
    String statement, ColumnValue[] cvs,
    int status) throws SQLException {

    StringBuffer sb = new StringBuffer(statement);
    int whereCount =
      status == Status.UPDATED ||
      status == Status.DELETED ?
      whereColumns.length : 0;

    Object[] values = new Object[whereCount];
    int[] types = new int[whereCount];

    if (status == Status.UPDATED ||
      status == Status.DELETED) {

      sb.append(" WHERE ");
      ReadRow oldRow = logEntry.getOldRow();

      for (int i = 0; i < whereCount; i++) {
        int index = getColumnIndex(whereColumns[i]);
        Column column = columnManager.getColumn(index);
        Format formatter = column.getFormatter();
        values[i] = Formatter.parseObject(
          formatter, oldRow.get(index));

        types[i] = column.getType();

        if (null == values[i]) {
          continue;
        }

        sb.append(whereColumns[i] +
          " = ");

        if (stringPrepare ||
          !JDBCUtil.isString(types[i])) {

          sb.append("?");
        } else {
          sb.append("'" + values[i] + "'");
        }

        sb.append(" AND ");
      }

      sb.setLength(sb.length() - 5);
    }

    if (traceSQL) {
      System.out.println(sb);
    }

    PreparedStatement updateStatement =
      database.prepareStatement(sb.toString());

    int count = cvs.length;
    int paramIndex = 0;
    for (int i = 0; i < count; i++) {
      int index = cvs[i].getColumnIndex();
      Column column = columnManager.getColumn(index);
      Format formatter = column.getFormatter();
      Object value = Formatter.parseObject(
        formatter, cvs[i].getValue());

      int type = column.getType();

      if (stringPrepare ||
        !JDBCUtil.isString(type)) {

        JDBCUtil.setInParameter(updateStatement,
          paramIndex, value, type);

        if (traceSQL) {
          System.out.println("parameter[" +
            (paramIndex + 1) + "]:" +
            Formatter.format(formatter, value));
        }

        paramIndex++;
      }
    }

    int mCount = status == Status.INSERTED &&
      masterDataSet != null ?
      masterColumnNames.length : 0;

    for (int i = 0; i < mCount; i++) {
      if (masterValues[i] == null) {
        continue;
      }

      if (stringPrepare ||
        !JDBCUtil.isString(masterTypes[i])) {

        JDBCUtil.setInParameter(
          updateStatement, paramIndex,
          masterValues[i], masterTypes[i]);

        if (traceSQL) {
          System.out.println("parameter[" +
            (paramIndex + 1) + "]:" +
            Formatter.format(null, masterValues[i]));
        }

        paramIndex++;
      }
    }

    for (int i = 0; i < whereCount; i++) {
      if (values[i] == null) {
        continue;
      }

      if (stringPrepare ||
        !JDBCUtil.isString(types[i])) {

        JDBCUtil.setInParameter(updateStatement,
          paramIndex, values[i], types[i]);

        if (traceSQL) {
          System.out.println("parameter[" +
            (paramIndex + 1) + "]:" +
            Formatter.format(null, values[i]));
        }

        paramIndex++;
      }
    }

    int updateCount = updateStatement.executeUpdate();
    if (updateCount != 1) {
      throw new SQLException(
        "updateCount:" + updateCount + "\n" + sb);
    }

    if (database.isAutoCommit()) {
      _commit(false);
    }
  }

  public void cancel() throws SQLException {
    int oldRowCount = getRowCount();
    _cancel();
    if (oldRowCount == getRowCount()) {
      fireDataChanged(Status.UPDATED);
    } else {
      fireDataChanged(Status.STRUCTURE_CHANGED);
    }

    fireMasterDataChanged();
    fireStatusMessage();
  }

  protected synchronized void _cancel() throws SQLException {
    if (logEntry == null) {
      return;
    }

    int status = logEntry.getStatus();
    int rowIndex = logEntry.getRowIndex();
    ReadRow oldRow = logEntry.getOldRow();

    switch (status) {
      case Status.INSERTED:
        __deleteRow(rowIndex);
        break;
      case Status.UPDATED:
        setRow(rowIndex, oldRow);
        break;
      case Status.DELETED:
        __insertRow(rowIndex);
        setRow(oldRow);
        break;
    }

    logEntry = null;
  }

  public void commit() throws SQLException {
    if (database != null) {
      commit(false);
    }

    fireStatusMessage();
  }

  public void commit(TransactionEvent e)
    throws SQLException {

    commit(true);
  }

  private synchronized void commit(boolean event)
    throws SQLException {

    if (!event) {
      _post();
    }

    _commit(event);
  }


  private synchronized void _commit(boolean event)
    throws SQLException {

    if (log.empty() || database == null) {
      return;
    }

    log = new Stack();

    if (!event) {
      logEntry = null;
      if (traceSQL) {
        System.out.println("COMMIT");
      }

      database.commit();
    }
  }

  public synchronized void rollback()
    throws SQLException {

    rollback(false);
  }

  public void rollback(TransactionEvent e)
    throws SQLException {

    rollback(true);
  }

  private synchronized void rollback(boolean event)
    throws SQLException {

    if (!event) {
      if (traceSQL) {
        System.out.println("ROLLBACK");
      }

      database.rollback();
    }

    if (masterColumnNames != null &&
      masterColumnNames.length > 0) {

      _cancel();
      log.removeAllElements();
      cashe.clear();
      oldMasterRowid = -1;
      return;
    }

    _cancel();
    while (!log.empty()) {
      logEntry = (LogEntry)log.pop();
      _cancel();
    }

    fireDataChanged(Status.STRUCTURE_CHANGED);
    fireMasterDataChanged();
    fireStatusMessage();
  }

  public Object get(String columnName) {
    int columnIndex = getColumnIndex(columnName);
    return get(columnIndex);
  }

  public Object get(int columnIndex) {
    return get(currentRowIndex, columnIndex);
  }

  public Object get(int rowIndex, int columnIndex) {
    ReadRow rr = getRow(rowIndex);
    return rr.get(columnIndex);
  }

  public String getString(String columnName) {
    int columnIndex = getColumnIndex(columnName);
    return getString(columnIndex);
  }

  public String getString(int columnIndex) {
    return getString(currentRowIndex, columnIndex);
  }

  public String getString(int rowIndex, int columnIndex) {
    Object value = get(rowIndex, columnIndex);
    Column column = columnManager.getColumn(columnIndex);
    Format formatter = column.getFormatter();

    return Formatter.format(formatter, value);
  }

  protected ReadWriteRow getRow() {
    return getRow(currentRowIndex);
  }

  protected ReadWriteRow getRow(int rowIndex) {
    if (rowIndex >= 0) {
      return (ReadWriteRow)rows.elementAt(rowIndex);
    } else {
      return new ReadWriteRow();
    }
  }

  public int getColumnCount() {
    return columnManager.getColumnCount();
  }

  public int getColumnIndex(String columnName) {
    return columnManager.getColumnIndex(columnName);
  }

  public String getColumnName(int columnIndex) {
    return columnManager.getColumnName(columnIndex);
  }

  public void setColumn(String columnName, Column column) {
    columnManager.setColumn(columnName, column);
  }

  public void setColumn(int columnIndex, Column column) {
    columnManager.setColumn(columnIndex, column);
  }

  public Column getColumn(String columnName) {
    return columnManager.getColumn(columnName);
  }

  public Column getColumn(int columnIndex) {
    return columnManager.getColumn(columnIndex);
  }

  public int getRowCount() {
    return rows.size();
  }

  public String getCatalogName() {
    return catalogName;
  }

  public String getSchemaName() {
    return schemaName;
  }

  public String getTableName() {
    return tableName;
  }

  public boolean next() throws SQLException {
    if (_next()) {
      fireNavigated();
      fireMasterDataChanged();
      fireStatusMessage();
      return true;
    } else {
      return false;
    }
  }

  protected synchronized boolean _next()
    throws SQLException {

    _post();
    return __next();
  }

  protected synchronized boolean __next()
    throws SQLException {

    if (isLast()) {
      return false;
    }

    currentRowIndex++;
    return true;
  }

  public boolean prior() throws SQLException {
    if (_prior()) {
      fireNavigated();
      fireMasterDataChanged();
      fireStatusMessage();
      return true;
    } else {
      return false;
    }
  }

  protected synchronized boolean _prior()
    throws SQLException {

    _post();
    return __prior();
  }

  protected synchronized boolean __prior() {
    if (isFirst()) {
      return false;
    }

    currentRowIndex--;
    return true;
  }

  public boolean first() throws SQLException {
    if (_first()) {
      fireNavigated();
      fireMasterDataChanged();
      fireStatusMessage();
      return true;
    } else {
      return false;
    }
  }

  protected synchronized boolean _first()
    throws SQLException {

    _post();
    fetch();

    return __first();
  }

  protected synchronized boolean __first() {
    if (isEmpty()) {
      return false;
    }

    currentRowIndex = 0;
    return true;
  }

  public boolean last() throws SQLException {
    if (_last()) {
      fireNavigated();
      fireMasterDataChanged();
      fireStatusMessage();
      return true;
    } else {
      return false;
    }
  }

  protected synchronized boolean _last()
    throws SQLException {

    _post();
    fetchAll();
    return __last();
  }

  protected synchronized boolean __last()
    throws SQLException {

    if (isEmpty()) {
      return false;
    }

    currentRowIndex = getRowCount() - 1;
    return true;
  }

  public boolean goToRow(int rowIndex) throws SQLException {
    if (_goToRow(rowIndex)) {
      fireNavigated();
      fireMasterDataChanged();
      fireStatusMessage();
      return true;
    } else {
      return false;
    }
  }

  protected synchronized boolean _goToRow(int rowIndex)
    throws SQLException {

    _post();
    return __goToRow(rowIndex);
  }

  protected synchronized boolean __goToRow(int rowIndex)
    throws SQLException {

    if (rowIndex < 0 || rowIndex + 1 > rows.size()) {
      return false;
    }

    currentRowIndex = rowIndex;
    return true;
  }

  public int getRowIndex() {
    return currentRowIndex;
  }

  public int getRowid() {
    ReadRow rr = getRow(currentRowIndex);
    return rr.getRowid();
  }

  public int getStatus() {
    if (logEntry == null) {
      return Status.NORMAL;
    } else {
      return logEntry.getStatus();
    }
  }

  public void addDataChangeListener(
    DataChangeListener listener) {

    dataChangeCaster.add(listener);
  }

  public void removeDataChangeListener(
    DataChangeListener listener) {

    dataChangeCaster.remove(listener);
  }

  public void addNavigationListener(
    NavigationListener listener) {

    navigationCaster.add(listener);
  }

  public void removeNavigationListener(
    NavigationListener listener) {

    navigationCaster.remove(listener);
  }

  public void addMasterDataChangeListener(
    MasterDataChangeListener listener) {

    masterDataChangeCaster.add(listener);
    if (isOpen()) {
      fireMasterDataChanged();
    }
  }

  public void removeMasterDataChangeListener(
    MasterDataChangeListener listener) {

    masterDataChangeCaster.remove(listener);
  }

  public void addStatusListener(
    StatusListener listener) {

    statusCaster.add(listener);
  }

  public void removeStatusListener(
    StatusListener listener) {

    statusCaster.remove(listener);
  }

  public void masterDataChanged(
    MasterDataChangeEvent e) throws SQLException {

    if (detailCashe &&
      oldMasterRowid >= 0) {

      HDataSet copy = new HDataSet(this);
      cashe.put(new Integer(oldMasterRowid), copy);
    }

    _post();
    __close();
    _masterDataChanged(e);
    _first();
    fireDataChanged(Status.STRUCTURE_CHANGED);
    fireNavigated();
    fireStatusMessage();
  }

  private synchronized void _masterDataChanged(
    MasterDataChangeEvent e) throws SQLException {

    HDataSet ds = null;

    oldMasterRowid = masterDataSet.getRowid();
    ds = (HDataSet)cashe.get(
      new Integer(oldMasterRowid));

    if (ds == null || ds.masterValues.length == 0) {
      int mCount = masterColumnNames.length;
      int dCount = detailColumnNames.length;

      if (dCount == 0) {
        detailColumnNames = new String[mCount];
        System.arraycopy(masterColumnNames, 0,
          detailColumnNames, 0, mCount);
      } else if (dCount != mCount) {
        throw new IllegalStateException(
          dCount + " != " + mCount);
      }

      masterValues = new Object[mCount];
      masterTypes = new int[mCount];

      StringBuffer sb = new StringBuffer(query);
      if (query.toUpperCase().indexOf(" WHERE ") > 0) {
        sb.append(" AND ");
      } else {
        sb.append(" WHERE ");
      }

      for (int i = 0; i < mCount; i++) {
        masterValues[i] =
          masterDataSet.get(masterColumnNames[i]);

        Column column = masterDataSet.getColumn(masterColumnNames[i]);
        masterTypes[i] = column.getType();

        if (masterValues[i] == null) {
          masterValues = new Object[0];
          masterTypes = new int[0];
          return;
        }

        sb.append(detailColumnNames[i] +
          " = ");

        if (stringPrepare ||
          !JDBCUtil.isString(masterTypes[i])) {

          sb.append("?");
        } else {
          sb.append("'" + masterValues[i] + "'");
        }

        sb.append(" AND ");
      }

      sb.setLength(sb.length() - 5);

      if (traceSQL) {
        System.out.println(sb);
      }

      selectStatement =
        database.prepareStatement(sb.toString());

      int pCount = parameters.size();
      for (int i = 0; i < mCount; i++) {
        if (stringPrepare ||
          !JDBCUtil.isString(masterTypes[i]) ||
          masterValues[i] != null) {

          if (traceSQL) {
            System.out.println("parameter[" +
              (pCount + i + 1) + "]:" +
              Formatter.format(null, masterValues[i]));
          }

          JDBCUtil.setInParameter(
            selectStatement, pCount + i,
            masterValues[i], masterTypes[i]);
        }
      }

      executeQuery();
      if (columnManager == null) {
        setupMetaData();
      }

      fetchAll();
    } else {
      restoreDataSet(ds);
      metaDataNeed = false;
      fetchNeed = false;
    }
  }

  private synchronized void setDataSet(HDataSet ds) {
    rows = (Vector)ds.rows.clone();
    currentRowIndex = ds.currentRowIndex;
    if (ds.columnManager != null) {
      columnManager = ds.columnManager;
    }

    masterValues = new Object[ds.masterValues.length];
    System.arraycopy(ds.masterValues, 0,
      masterValues, 0, ds.masterValues.length);

    masterTypes = new int[ds.masterTypes.length];
    System.arraycopy(ds.masterTypes, 0,
      masterTypes, 0, ds.masterTypes.length);
  }

  private synchronized void restoreDataSet(HDataSet ds) {
    setDataSet(ds);
    if (isRowEmpty() && currentRowIndex >= 0) {
      __deleteRow(currentRowIndex);
    }
  }

  public void setDetailCashe(boolean cashe) {
    this.detailCashe = cashe;
  }

  public boolean isDetailCashe() {
    return detailCashe;
  }

  public boolean isFirst() {
    return currentRowIndex <= 0;
  }

  public boolean isLast() {
    return currentRowIndex + 1 == getRowCount();
  }

  public boolean isTransaction() {
    return !log.empty() || isEditing();
  }

  public boolean isRowEmpty() {
    ReadRow rr = getRow();
    return rr.getColumnCount() == 0;
  }

  public boolean isEmpty() {
    return getRowCount() == 0;
  }

  public boolean isEditing() {
    return logEntry != null;
  }

  public String toString() {
    StringBuffer sb = new StringBuffer();

    try {
      if (!isOpen()) {
        open();
      }

      if (rows.size() == 0) {
        fetchAll();
      }

      int count = getColumnCount();
      int mCount = masterColumnNames.length;

      int[] displaySizes = new int[count];
      boolean[] numbers = new boolean[count];
      Format[] formatters = new Format[count];

      for (int i = 0; i < count; i++) {
        Column column = columnManager.getColumn(i);
        displaySizes[i] = column.getDisplaySize();
        numbers[i] = JDBCUtil.isNumber(
          column.getType());

        formatters[i] = column.getFormatter();

        String s = JDBCUtil.getDisplayString(
          getColumnName(i),
          displaySizes[i], numbers[i]);

        sb.append(s + " ");
      }

      sb.setCharAt(sb.length() - 1, '\n');
      sb.append(JDBCUtil.getSeparator(displaySizes));

      int rCount = rows.size();
      for (int r = 0; r < rCount; r++) {
        ReadRow rr = (ReadRow)rows.elementAt(r);
        for (int c = 0; c < count; c++) {
          String s = JDBCUtil.getDisplayString(
            Formatter.format(formatters[c], rr.get(c)),
            displaySizes[c], numbers[c]);

          sb.append(s + " ");
        }

        sb.setCharAt(sb.length() - 1, '\n');
      }
    } catch (Exception ex) {
      ex.printStackTrace();
    }

    return sb.toString();
  }

  protected synchronized void fetch()
    throws SQLException {

    if (!isFetchNeed()) {
      return;
    }

    int saveRowIndex = currentRowIndex;
    int count = getColumnCount();

    try {
      for (int r = 0; r < fetchSize; r++) {
        if (!resultSet.next()) {
          fetchNeed = false;
          break;
        }

        __insertRow(rows.size());
        ReadWriteRow rwr =
          (ReadWriteRow)rows.elementAt(currentRowIndex);
        for (int c = 0; c < count; c++) {
          Object value =
            resultSet.getObject(c + 1);

          rwr.set(c, value);
        }
      }
    } finally {
      currentRowIndex = saveRowIndex;
    }
  }

  protected void fetchAll() throws SQLException {
    int oldFetchSize = fetchSize;
    fetchSize = 10000;

    try {
      while (fetchNeed) {
        fetch();
      }
    } finally {
      fetchSize = oldFetchSize;
    }
  }

  private boolean isFetchNeed() {
    return fetchNeed &&
      currentRowIndex + 1 >= rows.size();
  }

  private void fireDataChanged() {
    fireDataChanged(getStatus());
  }

  private void fireDataChanged(int status) {
    if (!dataChangeCaster.hasListeners()) {
      return;
    }

    DataChangeEvent e =
      new DataChangeEvent(
      this, DataChangeEvent.DATA_CHANGED, status);

    dataChangeCaster.dispatch(e);
  }

  private void fireNavigated() {
    if (!navigationCaster.hasListeners()) {
      return;
    }

    NavigationEvent e =
      new NavigationEvent(
      this, NavigationEvent.NAVIGATED);

    navigationCaster.dispatch(e);
  }

  private void fireMasterDataChanged() {
    if (!masterDataChangeCaster.hasListeners()) {
      return;
    }

    MasterDataChangeEvent e =
      new MasterDataChangeEvent(
      this, MasterDataChangeEvent.MASTER_DATA_CHANGED);

    try {
      masterDataChangeCaster.exceptionDispatch(e);
    } catch (Exception ex) {
      ex.printStackTrace();
    }
  }

  private void fireStatusMessage() {
    if (!statusCaster.hasListeners()) {
      return;
    }

    StatusEvent e =
      new StatusEvent(
      this, StatusEvent.STATUS_MESSAGE);

    statusCaster.dispatch(e);
  }

  private synchronized void executeQuery()
    throws SQLException {

    registerParameter();
    resultSet = selectStatement.executeQuery();
    fetchNeed = true;
  }

  private synchronized void registerParameter()
    throws SQLException {

    int pCount = parameters.size();
    for (int i = 1; i <= pCount; i++) {
      Object value = parameters.get(
        new Integer(i));

      if (traceSQL) {
        System.out.println("parameter[" +
          i + "]:" + Formatter.format(null, value));
      }

      selectStatement.setObject(i , value);
    }
  }
}
